import numpy as np

from ieml.commons import cached_property
from ieml.dictionary.script import script
from .terms import Term
from .script import Script, factorize


class Table(Term):
    def __init__(self, script, index, dictionary, parent, regular=False):
        super().__init__(script, index, dictionary, parent)
        self.regular = regular

    @property
    def rank(self):
        if self.parent is None:
            return 0

        if isinstance(self.parent, TableSet):
            return max(self.parent.rank, 1)

        if self.regular:
            return max(self.parent.rank, 1) + 2
        else:
            return max(self.parent.rank, 1) + 1

    @cached_property
    def partitions(self):
        return {t for t in self.relations.contains if isinstance(t, Table) and t.parent == self}


class Cell(Term):
    @property
    def rank(self):
        return 6


class Table2D(Table):
    def __init__(self, script, index, dictionary, parent, regular=False):
        super().__init__(script, index, dictionary, parent, regular)

        if script.tables_script[0] != script or self.script.cells[0].shape[2] != 1 or self.script.cells[0].shape[1] == 1:
            raise ValueError("Invalid script for Table creation: %s. Expected a script that lead a 2d table"%str(script))

        self._index = None

    @property
    def shape(self):
        return self.cells.shape

    @cached_property
    def rows(self):
        return [self.dictionary.terms[row] if row in self.dictionary else None
                for row in self.script_rows]

    @cached_property
    def script_rows(self):
        return [factorize([t.script for t in line]) for line in self.cells]

    @cached_property
    def columns(self):
        return [self.dictionary.terms[column] if column in self.dictionary else None
                for column in self.script_columns]

    @cached_property
    def script_columns(self):
        return [factorize([t.script for t in line]) for line in self.cells.transpose()]

    @cached_property
    def cells(self):
        return np.vectorize(lambda sc: self.dictionary.terms[sc])(self.script.cells[0][:, :, 0])

    def __getitem__(self, item):
        return self.cells[item]

    def index_of(self, item):
        if self._index is None:
            self._index = {
                t.script: index for index, t in np.ndenumerate(self.cells)
            }

        return self._index[script(item)]

    def accept_script(self, script):
        """

        :param term:
        :return: is_accepting, is_regular
        """
        def is_connexe_tilling(coords):
            shape_t = self.shape
            # test if the table table is a connexe tiling of the table t
            size = 1
            for i, indexes in enumerate(zip(*coords)):

                indexes = sorted(set(indexes))
                size *= len(indexes)

                missing = [j for j in range(shape_t[i]) if j not in indexes]
                # more than one transition from missing -> included
                transitions = [j for j in missing if (j + 1) % shape_t[i] not in missing]
                if len(transitions) > 1:
                    step = int(shape_t[i] / len(transitions))

                    # check if there is a periodicity
                    if any((j + step)%shape_t[i] not in transitions for j in transitions):
                        return False

                # at least a square of 2x2x1
                if len(indexes) < 2 and i != 2:
                    return False

            if len(coords) == size:
                return True

            return False

        def is_dim_subset(coords):
            """
            Return if it is a dim subset (only one dimension, the others are not touched)
            If True, return (True, nb_dim)
            else (False, 0)

            :param coords:
            :param t:
            :return:
            """
            shape_ts0 = tuple(len(set(indexes)) for indexes in zip(*coords))
            shape_t = self.shape

            if shape_ts0[0] == shape_t[0]:
                return True, shape_ts0[1]

            if shape_ts0[1] == shape_t[1]:
                return True, shape_ts0[0]

            return False, 0

        if script not in self.script:
            return False, False

        if self.rank != 0 and self.rank % 2 == 0:
            return False, False

        coords = sorted([self.index_of(ss) for ss in script.singular_sequences])

        is_dim, count = is_dim_subset(coords)
        if is_dim and count == 1:
            # one dimension subset, it is a rank 3/5 paradigm
            return True, True

        # the coordinates sorted of the ss of s0 in the table t
        # rank is then 2/4
        if is_connexe_tilling(coords) or is_dim:
            return True, False

        return False, False


class Table1D(Table):
    def __init__(self, script, index, dictionary, parent, regular=False):
        super().__init__(script, index, dictionary, parent, regular)
        self._index = None

    @property
    def shape(self):
        return self.cells.shape

    @cached_property
    def cells(self):
        return np.vectorize(lambda sc: self.dictionary.terms[sc])(self.script.cells[0][:, 0, 0])

    def __getitem__(self, item):
        return self.cells[item]

    def index_of(self, item):
        if self._index is None:
            self._index = {
                t.script: index for index, t in np.ndenumerate(self.cells)
            }

        return self._index[script(item)]

    def accept_script(self, script):
        if script not in self.script:
            return False, False

        coords = sorted([self.index_of(ss) for ss in script.singular_sequences])

        if len(coords) == coords[-1][0] - coords[0][0] + 1:
            return True, True
        else:
            return False, False


class TableSet(Table):
    def __init__(self, script, index, dictionary, parent, regular=False):
        super().__init__(script, index, dictionary, parent, regular)

        if len(self.script.tables_script) == 1:
            raise ValueError("Invalid script for TableSet creation: %s. Expected a script that generate multiple "
                             "Tables"%str(script))

    @property
    def tables(self):
        return [self.dictionary.terms[s] for s in self.script.tables_script]

    def accept_script(self, script):
        """
        True when the term is a subset of this term tables. If the parent of this term is already a TableSet,return
        always false (only one main tableset)
        :param term:
        :return:
        """
        if isinstance(self.parent, TableSet):
            return False, False

        tables = [table for table in self.script.tables_script if table in script]

        if len(tables) >= 1 and {ss for t in tables for ss in t.singular_sequences} == set(script.singular_sequences):
            return True, False

        return False, False


class Table3D(TableSet):
    def __init__(self, script, index, dictionary, parent, regular=False):
        super().__init__(script, index, dictionary, parent, regular)

    def __getitem__(self, item):
        return self.tables[item[0]][item[1:]]


def table_class(script):
    if len(script) == 1:
        return Cell

    if len(script.tables_script) == 1:
        dim = sum(1 for s in script.cells[0].shape if s != 1)
        if dim == 1:
            return Table1D
        if dim == 2:
            return Table2D

        raise ValueError("Invalid dim %d for script %s"%(dim, str(script)))
    else:
        if len(script.cells) == 1:
            return Table3D
        else:
            return TableSet



from enum import Enum
from ieml.calculation.distance import (paradigmatic_equivalence_class_index, set_proximity_index,
                                       object_proximity_index, connexity_index, mutual_inclusion_index)
from math import ceil

from ieml.ieml_objects.hypertexts import Hypertext
from ieml.ieml_objects.sentences import Sentence, SuperSentence
from ieml.ieml_objects.terms import Term
from ieml.ieml_objects.texts import Text
from ieml.ieml_objects.words import Word
from ieml.usl.tools import usl
from ieml.usl.usl import Usl


class FilteringLevel(Enum):
    UNITERM_WORD = 1
    MULTITERM_WORD = 2
    SENTENCE = 3
    SUPERSENTENCE = 4

    @classmethod
    def get_usl_filtering_level(cls, input_usl):
        """If the max level for a USL is a sentence or supersentence, then its filtering level is the same,
        else, this function figures out if it's a multiterm on uniterm word"""

        input_usl = usl(input_usl)

        usl_max_level = input_usl.max_level
        # TODO : unittest this function
        if usl_max_level == Sentence:
            return cls.SENTENCE
        elif usl_max_level == SuperSentence:
            return cls.SUPERSENTENCE
        else: # must be a word, we have to figure out if single term or not
            # all of the USL's elements have to be words, so for each words, we check if the substance's
            # count is 1 and if the mode is empty
            for word in usl:
                if len(word.subst.children) != 1 or word.mode is not None:
                    return cls.MULTITERM_WORD
            return cls.UNITERM_WORD # reached only if all words are monoterm

    @classmethod
    def get_proposition_filtering_level(cls, proposition):
        if isinstance(proposition, (Sentence, SuperSentence)):
            return cls.SENTENCE if isinstance(proposition, Sentence) else cls.SUPERSENTENCE
        else:
            if len(proposition.subst.children) != 1 or proposition.mode is not None:
                return cls.MULTITERM_WORD
            else:
                return cls.UNITERM_WORD

    @classmethod
    def get_higher_levels(cls, level):
        return [cls(index) for index in range(level.value, cls.SUPERSENTENCE.value + 1)]


class AbstractFilter:
    def __init__(self, level=Term, ratio=None):
        self.level = level
        self.ratio = ratio

    def filter(self, query_usl, usl_list, ratio):
        """Returns a list of filtered usl, using the ratio"""
        pass


class ParadigmaticProximityFilter(AbstractFilter):
    """Filter based on the P(OE^1) indicator"""

    def filter(self, query_usl, usl_list, ratio):
        """Overrides the AbstractFilter method to return a list ranked based on the
        P(OE^1) indicator """

        # TODO: check which paradigm table rank to use and how (i.e. should we take the mean of the indicator over all ranks)
        usl_score = {usl: paradigmatic_equivalence_class_index(query_usl, usl, 1, "OE") for usl in usl_list}
        sorted_by_score = sorted(usl_score, key=lambda e: usl_score[e], reverse=True)
        return sorted_by_score[:ceil(ratio*len(usl_list))], usl_score

    def __str__(self):
        return "P(OE^1)"


class IndicatorFilter(AbstractFilter):
    """Filter based on any indicator, parametrized by the indicator function."""

    def __init__(self, indicator_function, **kwargs):
        super().__init__(**kwargs)
        self.indicator = indicator_function

    def filter(self, query_usl, usl_list, ratio):
        usl_score = {usl: self.indicator(self.level, query_usl, usl) for usl in usl_list}
        sorted_by_score = sorted(usl_score, key=lambda e: usl_score[e], reverse=True)
        return sorted_by_score[:ceil(ratio * len(usl_list))], usl_score

    def __str__(self):
        function_to_symbol_mapping = {
            set_proximity_index : "EO",
            object_proximity_index : "(O,O)",
            connexity_index : "(O-O)",
            mutual_inclusion_index: "(O,o)"
        }
        return function_to_symbol_mapping[self.indicator] + " lvl %s" % self.level.__name__


class BinaryFilter:

    def __init__(self, mode=FilteringLevel.MULTITERM_WORD):
        self.filtering_level = mode
        self.mode = type_mapping[mode]

    def __str__(self):
        return "BinF(%s)" %  self.mode.__name__

    def _filter_by_lvl(self, propositions_list):
        return {proposition for proposition in propositions_list
                if FilteringLevel.get_proposition_filtering_level(proposition) == self.filtering_level}

    def filter(self, query_usl, usl_list):
        """Filters out the USL who's propositions of the filter's filteringlevel do not intersect with the queries's
        proposition of the same filteringlevel"""
        query_usl_propositions = self._filter_by_lvl(query_usl.texts[0].children)
        return [usl for usl in usl_list
                if query_usl_propositions.intersection(self._filter_by_lvl(usl.texts[0].children))]

type_mapping = {
    FilteringLevel.MULTITERM_WORD: Word,
    FilteringLevel.SENTENCE: Sentence,
    FilteringLevel.SUPERSENTENCE: SuperSentence
}




class PathError(Exception):
    def __init__(self, message, path):
        self.message = message
        self.path = path

    def __str__(self):
        return self.message + "[%s]"%str(self.path)


class ResolveError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return str(self.message)


class IEMLObjectResolutionError(Exception):
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return "Invalid ieml object definitions at:\n%s"%('\n'.join("\t[%s] %s"%(str(p), e) for p, e in self.errors))


from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        return f.read()


setup(name='ieml',
      version='0.2.43',
      description='Implementation of the artificial natural language IEML',
      long_description=readme(),
      classifiers=[
            'Development Status :: 3 - Alpha',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Programming Language :: Python :: 3.5',
            'Topic :: Text Processing :: Linguistic',
            'Topic :: Text Processing :: Indexing'
      ],
      keywords='ieml semantic syntax relations',
      url='https://github.com/IEMLdev/ieml',
      author='Louis van Beurden',
      author_email='louis.vanbeurden@gmail.com',
      license='GPLv3',
      packages=find_packages(exclude=['scripts', '*.test']),
      install_requires=[
            'numpy',
            'bidict',
            'ply',
            'scipy',
      ],
      test_suite='nose2.collector.collector',
      tests_require=['nose2'],

      include_package_data=True,
      zip_safe=False)


import logging
from configparser import ExtendedInterpolation
import os
from os.path import isfile, isdir, expanduser, join, dirname
import configparser
import sys


_config = configparser.ConfigParser(interpolation=ExtendedInterpolation())
_config.read(join(dirname(__file__), 'default_config.conf'))

_HOME_FOLDER = expanduser("~")
ieml_folder = join(_HOME_FOLDER, _config.get('DEFAULT', 'iemlfolder'))
_config_file = join(ieml_folder, _config.get('DEFAULT', 'configfile'))
parser_folder = join(ieml_folder, 'parser')


def init_logging(config):
    level = getattr(logging, config.get('DEFAULT', 'loglevel').upper())
    if not isinstance(level, int):
        raise ValueError('Invalid log level: %s' % level)

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    try:
        file = config.get('DEFAULT', 'logfile')
    except configparser.NoOptionError:
        pass
    else:
        fh = logging.FileHandler(file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        root.addHandler(fh)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(level)
    ch.setFormatter(formatter)
    root.addHandler(ch)


if not isdir(ieml_folder):
    os.mkdir(ieml_folder)

if not isdir(parser_folder):
    os.mkdir(parser_folder)

if isfile(_config_file):
    _config.read(_config_file)

init_logging(_config)

def get_configuration():
    return _config


from . import tools
from . import exceptions
from . import constants
from . import syntax
from . import dictionary
from . import usl




from ieml.dictionary.terms import Term
from ieml.syntax.terms import SyntaxTerm
from .commons import IEMLSyntax
from ..exceptions import InvalidIEMLObjectArgument
from ..constants import MAX_SINGULAR_SEQUENCES, MORPHEME_SIZE_LIMIT


class Morpheme(IEMLSyntax):
    def __init__(self, children):
        if isinstance(children, SyntaxTerm):
            _children = (children,)
        else:
            try:
                _children = tuple(e for e in children)
            except TypeError:
                raise InvalidIEMLObjectArgument(Morpheme, "The argument %s is not an iterable" % str(children))

        if not 0 < len(_children) <= MORPHEME_SIZE_LIMIT:
            raise InvalidIEMLObjectArgument(Morpheme, "Invalid terms count %d."%len(_children))

        if not all((isinstance(e, SyntaxTerm) for e in _children)):
            raise InvalidIEMLObjectArgument(Morpheme, "%s do not contain only SyntaxTerm instances."%(str(_children)))

        # check singular sequences intersection
        singular_sequences = [s for t in _children for s in t.script.singular_sequences]
        if len(singular_sequences) != len(set(singular_sequences)):
            raise InvalidIEMLObjectArgument(Morpheme, "Singular sequences intersection in %s."%
                                            str([str(t) for t in _children]))

        super().__init__(sorted(_children))

        self.cardinal = 1
        for c in self.children:
            self.cardinal *= c.script.cardinal

    @property
    def grammatical_class(self):
        return max(t.grammatical_class for t in self.children)

    @property
    def empty(self):
        return len(self.children) == 1 and self.children[0].empty

    def compute_str(self, children_str):
        return '('+'+'.join(children_str)+')'


class Word(IEMLSyntax):
    closable = True

    def __init__(self, root=None, flexing=None, literals=None, children=None):

        if root is not None:
            _children = (root,) + ((flexing,) if flexing else ())
        else:
            _children = tuple(children)

        for c in _children:
            if not isinstance(c, Morpheme):
                raise InvalidIEMLObjectArgument(Word,
                                            "The children %s of a word must be a Morpheme instance." % (str(c)))

        # the root of a word can't be empty
        if _children[0].empty:
            raise InvalidIEMLObjectArgument(Word, "The root of a Word cannot be empty (%s)."%str(_children[0]))

        super().__init__(_children, literals=literals)

        self.cardinal = self.root.cardinal * (self.flexing.cardinal if self.flexing else 1)
        if self.cardinal > MAX_SINGULAR_SEQUENCES:
            raise InvalidIEMLObjectArgument(Word, "Too many word- singular sequences defined (max: 360) here: %d"%self.cardinal)

    @property
    def grammatical_class(self):
        return self.root.grammatical_class

    @property
    def root(self):
        return self.children[0]

    @property
    def flexing(self):
        if len(self.children) > 1:
            return self.children[1]
        return []

    def compute_str(self, children_str):
        return '['+'*'.join(children_str)+']'

    @property
    def is_term(self):
        return len(self.root.children) == 1 and not self.flexing

    @staticmethod
    def from_term(term):
        if isinstance(term, Term):
            term = SyntaxTerm(term)

        return Word(root=Morpheme(term))


import logging
from collections import defaultdict
from ieml.commons import cached_property
from ieml.dictionary.relations import RelationsGraph
from ieml.dictionary.table import Cell, table_class
from ieml.dictionary.version import save_dictionary_to_cache, load_dictionary_from_cache
from ieml.exceptions import TermNotFoundInDictionary, ScriptNotDefinedInVersion
from .version import DictionaryVersion, get_default_dictionary_version
from ..constants import MAX_LAYER
from .script import script
import threading
from ieml import get_configuration
import gc

USE_CACHE = get_configuration().get("RELATIONS", "cacherelations")
logger = logging.getLogger(__name__)


class DictionarySingleton(type):
    _instance = None

    # Forbid multiple dictionary creation in parallel
    lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if len(args) < 1 or not isinstance(args[0], (DictionaryVersion, str)):
            version = get_default_dictionary_version()
        elif isinstance(args[0], DictionaryVersion):
            version = args[0]
        elif isinstance(args[0], str):
            version = DictionaryVersion(args[0])
        else:
            raise ValueError("Invalid argument for dictionary creation, expected dictionary version, not %s"%str(args[0]))

        with cls.lock:
            if cls._instance is None or cls._instance.version != version:

                if cls._instance is not None:
                    del cls._instance
                    gc.collect()

                # check cache
                if not version.is_cached or not USE_CACHE:
                    cls._instance = super(DictionarySingleton, cls).__call__(version, **kwargs)

                    if USE_CACHE:
                        save_dictionary_to_cache(cls._instance)
                else:
                    cls._instance = load_dictionary_from_cache(version)

        return cls._instance


class Dictionary(metaclass=DictionarySingleton):
    def __init__(self, version=None):
        super().__init__()

        if isinstance(version, str):
            version = DictionaryVersion(version)

        self.version = version

        # buildable elements
        self.scripts = None
        self.terms = None
        self.roots = None
        self.inhibitions = None
        self.index = None
        self.relations_graph = None

        self._populate()
        logger.log(logging.INFO, "Dictionary loaded (version: %s, nb_roots: %d, nb_terms: %d)"%
              (str(self.version), len(self.roots), len(self)))

    @cached_property
    def translations(self):
        return self.version.translations

    @cached_property
    def layers(self):
        _layers = [[] for _ in range(MAX_LAYER + 1)]
        for t in self.index:
            _layers[t.script.layer].append(t)

        return _layers

    def __len__(self):
        return len(self.terms)

    def __contains__(self, item):
        return script(item) in self.terms

    def __iter__(self):
        return self.index.__iter__()

    def _define_root(self, root, paradigms, script_index):
        paradigms = sorted(paradigms, key=len, reverse=True)

        self.terms[root] = table_class(root)(root, index=script_index[root], dictionary=self, parent=None)
        defined = {self.terms[root]}

        for ss in root.singular_sequences:
            self.terms[ss] = Cell(script=ss, index=script_index[ss], dictionary=self, parent=self.terms[root])

        for s in paradigms:
            if s in self.terms:
                continue

            candidates = set()
            for t in defined:

                accept, regular = t.accept_script(s)
                if accept:
                    candidates |= {(t, regular)}

            if len(candidates) == 0:
                raise ValueError("No parent candidate for the table produced by term %s" % str(s))

            if len(candidates) > 1:
                logger.log(logging.ERROR, "Multiple parent candidate for the table produced by script %s: {%s} "
                      "choosing the smaller one." % (str(s), ', '.join([str(c[0]) for c in candidates])))

            parent, regular = min(candidates, key=lambda t: t[0])

            self.terms[s] = table_class(s)(script=s,
                                           index=script_index[s],
                                           dictionary=self,
                                           parent=parent,
                                           regular=regular)
            defined.add(self.terms[s])

        self.roots[self.terms[root]] = sorted(defined | set(self.terms[root]))

    def _populate(self, scripts=None, relations=None):
        self.version.load()

        if scripts is None:
            self.scripts = sorted(script(s) for s in self.version.terms)
        else:
            self.scripts = scripts

        script_index = {
            s: i for i, s in enumerate(self.scripts)
        }
        roots = defaultdict(list)
        root_ss = {}
        for root in self.version.roots:
            root = script(root)
            for ss in root.singular_sequences:
                root_ss[ss] = root

        for s in self.scripts:
            if s.cardinal == 1:
                continue

            roots[root_ss[s.singular_sequences[0]]].append(s)

        self.terms = {}
        self.roots = {}
        for root in self.version.roots:
            self._define_root(root=script(root), paradigms=roots[root], script_index=script_index)

        self.index = sorted(self.terms.values())
        self.inhibitions = {self.terms[r]: relations_list for r, relations_list in self.version.inhibitions.items()}

        if relations is None:
            self.relations_graph = RelationsGraph(dictionary=self)
        else:
            assert all(rel.shape[0] == len(self) and rel.shape[1] == len(self) for rel in relations)
            self.relations_graph = relations

    def __getstate__(self):
        return {
            'relations': self.relations_graph,
            'scripts': self.scripts,
            'version': str(self.version)
        }

    def __setstate__(self, state):
        if isinstance(state['version'], DictionaryVersion):
            self.version = state['version']
        else:
            self.version = DictionaryVersion(state['version'])
        self._populate(scripts=state['scripts'], relations=state['relations'])

    def translate_script_from_version(self, version, old_script):
        diff = self.version.diff_for_version(version)

        try:
            new_script = script(diff[old_script])
        except KeyError:
            raise ScriptNotDefinedInVersion(old_script, version)

        try:
            return self.terms[new_script]
        except KeyError:
            raise TermNotFoundInDictionary(new_script, self)


from .parser import ScriptParser
from .script import MultiplicativeScript, Script
from .tools import factorize


def m(substance, attribute=None, mode=None):
    children = (substance, attribute, mode)
    if all(isinstance(s, (Script, None.__class__)) for s in children):
        return MultiplicativeScript(children=children)
    else:
        raise NotImplemented


def script(arg):
    from ..terms import Term

    if isinstance(arg, str):
        s = ScriptParser().parse(arg)
        return s
    elif isinstance(arg, Script):
        return arg
    elif isinstance(arg, Term):
        return arg.script
    else:
        try:
            arg = [script(a) for a in arg]
            return factorize(arg)
        except TypeError:
            pass

    raise NotImplemented


# shorthand
sc = script



import logging
import os
from functools import partial
import ply.yacc as yacc

from ieml.dictionary.terms import Term
from ieml.dictionary.version import DictionaryVersion
from ieml.dictionary.dictionary import Dictionary
from ieml.dictionary.tools import term
from ieml.exceptions import TermNotFoundInDictionary, InvalidIEMLObjectArgument
from ieml.syntax.terms import SyntaxTerm
from ... import parser_folder
from ...exceptions import CannotParse
from ieml.syntax import Word, Morpheme, Clause, SuperClause, Sentence, SuperSentence, Text, Hypertext, Hyperlink, PropositionPath

from .lexer import get_lexer, tokens
import threading

def _add(lp1, p2):
    return lp1[0] + [p2[0]], lp1[1] + p2[1]


def _build(*arg):
    if len(arg) > 1:
        _h = []
        for e in arg[1:]:
            for h in e[1]:
                h[1].insert(0, arg[0])
                _h.append(h)
        return arg[0], _h
    return arg[0], []


def _hyperlink(node, text_list):
    return node[0], node[1] + [(text, [node[0]]) for text in text_list[0]]


class IEMLParserSingleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        dictionary = args[0] if len(args) > 0 else \
            kwargs['dictionary'] if 'dictionary' in kwargs else None

        if dictionary is None:
            dictionary = Dictionary()

        if not isinstance(dictionary, Dictionary):
            dictionary = Dictionary(dictionary)

        if dictionary.version not in cls._instances:
            # this code is to clean up duplicate class if we reload modules
            cls._instances[dictionary.version] = \
                super(IEMLParserSingleton, cls).__call__(dictionary=dictionary)

        return cls._instances[dictionary.version]


class IEMLParser(metaclass=IEMLParserSingleton):
    tokens = tokens
    lock = threading.Lock()

    def __init__(self, dictionary=None):

        self._get_term = partial(term, dictionary=dictionary)

        # Build the lexer and parser
        self.lexer = get_lexer()
        self.parser = yacc.yacc(module=self, errorlog=logging, start='proposition',
                                debug=False, optimize=True, picklefile=os.path.join(parser_folder, "ieml_parser.pickle"))
        self._ieml = None

    def parse(self, s):
        """Parses the input string, and returns a reference to the created AST's root"""
        # self._ieml = s
        # self.root = None
        # self.hyperlinks = []
        with self.lock:
            try:
                return self.parser.parse(s, lexer=self.lexer)
            except InvalidIEMLObjectArgument as e:
                raise CannotParse(s, str(e))
            except CannotParse as e:
                e.s = s
                raise e

            # if self.root is not None:
        #     if self.hyperlinks:
        #         self.root = Hypertext(self.hyperlinks)
        #
        #     # if isinstance(self.root, Term):
        #     #     self.root = Word(root=Morpheme([self.root]))
        #
        #     return self.root
        # else:
        #     raise CannotParse(s, "Invalid ieml object.")

    # Parsing rules
    def p_ieml_proposition(self, p):
        """proposition : script
                        | term
                        | morpheme
                        | word
                        | clause
                        | sentence
                        | superclause
                        | supersentence
                        | text"""
        if isinstance(p[1], Term):
            # script rule
            p[0] = SyntaxTerm(p[1])
        else:
            p[0] = p[1][0]

    def p_literal_list(self, p):
        """literal_list : literal_list LITERAL
                        | LITERAL"""

        if len(p) == 3:
            p[0] = p[1] + [p[2][1:-1]]
        else:
            p[0] = [p[1][1:-1]]


    def p_script(self, p):
        """script : TERM """
        try:
            p[0] = self._get_term(p[1])
        except TermNotFoundInDictionary as e:
            raise CannotParse(self._ieml, str(e))

    def p_term(self, p):
        """term : LBRACKET script RBRACKET
                | LBRACKET script RBRACKET literal_list"""
        if len(p) == 4:
            p[0] = _build(SyntaxTerm(p[2]))
        else:
            p[0] = _build(SyntaxTerm(p[2], literals=p[4]))


    def p_proposition_sum(self, p):
        """terms_sum : terms_sum PLUS term
                    | term
            clauses_sum : clauses_sum PLUS clause
                    | clause
            superclauses_sum : superclauses_sum PLUS superclause
                    | superclause
            closed_proposition_list : closed_proposition_list closed_proposition
                                    | closed_proposition"""
        if len(p) == 4:
            p[0] = _add(p[1], p[3])
        elif len(p) == 3:
            p[0] = _add(p[1], p[2])
        else:
            p[0] = _add(([], []), p[1])

    def p_morpheme(self, p):
        """morpheme : LPAREN terms_sum RPAREN"""
        p[0] = _build(Morpheme(p[2][0]), p[2])

    def p_word(self, p):
        """word : LBRACKET morpheme RBRACKET
                | LBRACKET morpheme RBRACKET literal_list
                | LBRACKET morpheme TIMES morpheme RBRACKET
                | LBRACKET morpheme TIMES morpheme RBRACKET literal_list"""

        if len(p) == 4:
            p[0] = _build(Word(root=p[2][0]), p[2])
        elif len(p) == 5:
            p[0] = _build(Word(root=p[2][0], literals=p[4]), p[2])
        elif len(p) == 6:
            p[0] = _build(Word(root=p[2][0], flexing=p[4][0]), p[2], p[4])
        else:
            p[0] = _build(Word(root=p[2][0], flexing=p[4][0], literals=p[6]), p[2], p[4])

    def p_decorated(self, p):
        """decorated_word : word
                           | word text_list
            decorated_sentence : sentence
                                | sentence text_list
            decorated_supersentence : supersentence
                                    | supersentence text_list"""
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = _hyperlink(p[1], p[2])

    def p_clause(self, p):
        """clause : LPAREN decorated_word TIMES decorated_word TIMES decorated_word RPAREN"""
        p[0] = _build(Clause(substance=p[2][0], attribute=p[4][0], mode=p[6][0]), p[2], p[4], p[6])

    def p_sentence(self, p):
        """sentence : LBRACKET clauses_sum RBRACKET
                    | LBRACKET clauses_sum RBRACKET literal_list"""
        if len(p) == 4:
            p[0] = _build(Sentence(p[2][0]), p[2])
        else:
            p[0] = _build(Sentence(p[2][0], literals=p[4]), p[2])

    def p_superclause(self, p):
        """superclause : LPAREN decorated_sentence TIMES decorated_sentence TIMES decorated_sentence RPAREN"""
        p[0] = _build(SuperClause(substance=p[2][0], attribute=p[4][0], mode=p[6][0]), p[2], p[4], p[6])

    def p_super_sentence(self, p):
        """supersentence : LBRACKET superclauses_sum RBRACKET
                         | LBRACKET superclauses_sum RBRACKET literal_list"""
        if len(p) == 4:
            p[0] = _build(SuperSentence(p[2][0]), p[2])
        else:
            p[0] = _build(SuperSentence(p[2][0], literals=p[4]), p[2])

    def p_closed_proposition(self, p):
        """ closed_proposition : SLASH decorated_word SLASH
                               | SLASH decorated_sentence SLASH
                               | SLASH decorated_supersentence SLASH"""
        p[0] = p[2]

    def p_text(self, p):
        """text : L_CURLY_BRACKET closed_proposition_list R_CURLY_BRACKET"""
        p[0] = _build(Text(p[2][0]))

        if p[2][1]:
            raise NotImplementedError("Ieml doesn't support hypertext parsing for the moment.")
            # tuple of the hyperlink (end text, path)
            # self.hyperlinks += [Hyperlink(p[0][0], e[0], PropositionPath(e[1])) for e in p[2][1]]

    def p_text_list(self, p):
        """text_list : text_list text
                    | text"""
        if len(p) == 3:
            p[0] = _add(p[1], p[2])
        else:
            p[0] = _add(([], []), p[1])

    def p_error(self, p):
        if p:
            msg = "Syntax error at '%s' (%d, %d)" % (p.value, p.lineno, p.lexpos)
        else:
            msg = "Syntax error at EOF"

        raise CannotParse(None, msg)



import xml.etree.ElementTree as ET
import random
import itertools
import functools

from urllib.request import urlopen

from ieml.exceptions import CannotParse

from ieml.syntax.parser.parser import IEMLParser
from ieml.syntax.commons import IEMLSyntax
from ieml.syntax.terms import SyntaxTerm
from ieml.dictionary.version import get_default_dictionary_version
from .exceptions import InvalidIEMLObjectArgument
from .syntax import Sentence, Clause, SuperSentence, SuperClause, Text, Word, Morpheme
from .exceptions import CantGenerateElement
from .dictionary import Term, Dictionary



def _loop_result(max_try):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ex = None
            for i in range(max_try):
                try:
                    return func(*args, **kwargs)
                except InvalidIEMLObjectArgument as e:
                    ex = e
                    continue

            raise CantGenerateElement(str(ex))
        return wrapper
    return decorator


class RandomPoolIEMLObjectGenerator:
    def __init__(self, level=Text, pool_size=20, dictionary_version=None):
        self.level = level
        self.pool_size = pool_size
        self.dictionary_version = dictionary_version

        if level > Text:
            raise ValueError('Cannot generate object higher than a Text.')

        self._build_pools()

        self.type_to_method = {
            Term: self.term,
            Word: self.word,
            Sentence: self.sentence,
            SuperSentence: self.super_sentence,
            Text: self.text
        }

    def _build_pools(self):
        """
        Slow method, retrieve all the terms from the database.
        :return:
        """
        if self.level >= Word:
            # words
            self.words_pool = set(self.word() for i in range(self.pool_size))

        if self.level >= Sentence:
            # sentences
            self.sentences_pool = set(self.sentence() for i in range(self.pool_size))

        if self.level >= SuperSentence:
            self.super_sentences_pool = set(self.super_sentence() for i in range(self.pool_size))

        if self.level >= Text:
            self.propositions_pool = set(itertools.chain.from_iterable((self.words_pool, self.sentences_pool, self.super_sentences_pool)))

        # self.hypertext_pool = set(self.hypertext() for i in range(self.pool_size))

    @_loop_result(10)
    def term(self):
        return SyntaxTerm(random.sample(Dictionary(self.dictionary_version).index, 1)[0])

    @_loop_result(10)
    def uniterm_word(self):
        return Word(Morpheme(SyntaxTerm(random.sample(Dictionary(self.dictionary_version).index, 1))))

    @_loop_result(10)
    def word(self):
        return Word(Morpheme([SyntaxTerm(t) for t in random.sample(Dictionary(self.dictionary_version).index, 3)]),
                    Morpheme([SyntaxTerm(t) for t in random.sample(Dictionary(self.dictionary_version).index, 2)]))

    def _build_graph_object(self, primitive, mode, object, max_nodes=6):
        nodes = {primitive()}
        modes = set()

        if max_nodes < 2:
            raise ValueError('Max nodes >= 2.')

        result = set()

        for i in range(random.randint(2, max_nodes)):
            while True:
                s, a, m = random.sample(nodes, 1)[0], primitive(), mode()
                if a in nodes or m in nodes or a in modes:
                    continue

                nodes.add(a)
                modes.add(m)

                result.add(object(s, a, m))
                break
        return result

    @_loop_result(10)
    def sentence(self, max_clause=6):
        def p():
            return random.sample(self.words_pool, 1)[0]

        return Sentence(self._build_graph_object(p, p, Clause, max_nodes=max_clause))

    @_loop_result(10)
    def super_sentence(self, max_clause=4):
        def p():
            return random.sample(self.sentences_pool, 1)[0]

        return SuperSentence(self._build_graph_object(p, p, SuperClause, max_nodes=max_clause))

    @_loop_result(10)
    def text(self):
        return Text(random.sample(self.propositions_pool, random.randint(1, 8)))

    def from_type(self, type):
        try:
            return self.type_to_method[type]()
        except KeyError:
            raise ValueError("Can't generate that type or not an IEMLObject : %s"%str(type))


def list_bucket(url):
    root_node = ET.fromstring(urlopen(url).read())
    all_versions_entry = ({k.tag: k.text for k in list(t)} for t in root_node
                          if t.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Contents')

    # sort by date
    all_versions = sorted(all_versions_entry,
                          key=lambda t: t['{http://s3.amazonaws.com/doc/2006-03-01/}LastModified'], reverse=True)

    return [v['{http://s3.amazonaws.com/doc/2006-03-01/}Key'] for v in all_versions]


def ieml(arg, dictionary_version=None):
    if not dictionary_version:
        dictionary_version = get_default_dictionary_version()

    if isinstance(arg, IEMLSyntax):
        if arg.dictionary_version != dictionary_version:
            arg.set_dictionary_version(dictionary_version)

        return arg

    if isinstance(arg, str):
        try:
            return IEMLParser(Dictionary(dictionary_version)).parse(arg)
        except CannotParse as e:
            raise InvalidIEMLObjectArgument(IEMLSyntax, str(e))

    if isinstance(arg, Term):
        arg = SyntaxTerm(arg)
        if arg.dictionary_version != dictionary_version:
            arg.set_dictionary_version(dictionary_version)

        return arg

    raise NotImplemented